﻿#!/usr/bin/env python
# -*- Python -*-

import sys
import time
sys.path.append(".")

# Import RTM module
import OpenRTM
import RTC
import ode  
import os
import codecs
from numpy import *
from math import *

# Import Service implementation class
# <rtc-template block="service_impl">

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
# </rtc-template>


# This module's spesification
# <rtc-template block="module_spec">
pySimu_spec = ["implementation_id", "pySimu", 
		 "type_name",         "pySimu", 
		 "description",       "py Simu Component", 
		 "version",           "0.1", 
		 "vendor",            "Kyouzyu", 
		 "category",          "TEST", 
		 "activity_type",     "DataFlowComponent", 
		 "max_instance",      "1", 
		 "language",          "Python", 
		 "lang_type",         "SCRIPT",
		 ""]
# </rtc-template>

class pySimu(OpenRTM.DataFlowComponentBase):
	def __init__(self, manager):
		OpenRTM.DataFlowComponentBase.__init__(self, manager)

		self._d_Torque = RTC.TimedDoubleSeq(RTC.Time(0,0),[])
		self._TorqueIn = OpenRTM.InPort("Torque", self._d_Torque, OpenRTM.RingBuffer(8))
		self._d_DesFlag = RTC.TimedBoolean(RTC.Time(0,0),0)
		self._DesFlagIn = OpenRTM.InPort("DesFlag", self._d_DesFlag, OpenRTM.RingBuffer(8))
		self._d_Pos = RTC.TimedDoubleSeq(RTC.Time(0,0),[])
		self._PosOut = OpenRTM.OutPort("Pos", self._d_Pos, OpenRTM.RingBuffer(8))
		self._d_Rot = RTC.TimedDoubleSeq(RTC.Time(0,0),[])
		self._RotOut = OpenRTM.OutPort("Rot", self._d_Rot, OpenRTM.RingBuffer(8))
		self._d_Angle = RTC.TimedDoubleSeq(RTC.Time(0,0),[])
		self._AngleOut = OpenRTM.OutPort("Angle", self._d_Angle, OpenRTM.RingBuffer(8))
		self._d_Pre = RTC.TimedDoubleSeq(RTC.Time(0,0),[])
		self._PreOut = OpenRTM.OutPort("Pre", self._d_Pre, OpenRTM.RingBuffer(8))
		

		# Set InPort buffers
		self.registerInPort("Torque",self._TorqueIn)
		self.registerInPort("DesFlag",self._DesFlagIn)
		self.step = 0
		
		# Set OutPort buffers
		self.registerOutPort("Rot",self._RotOut)
		self.registerOutPort("Pos",self._PosOut)
		self.registerOutPort("Angle",self._AngleOut)
		self.registerOutPort("Pre",self._PreOut)

		self.Simulater = Simu()
		

		

		

		
		# Set service provider to Ports
		
		# Set service consumers to Ports
		
		# Set CORBA Service Ports
		

		# initialize of configuration-data.
		# <rtc-template block="init_conf_param">
		
		# </rtc-template>


		 
	def onInitialize(self):
		# Bind variables and configuration variable
		
		return RTC.RTC_OK


	
	#def onFinalize(self, ec_id):
	#
	#	return RTC.RTC_OK
	
	#def onStartup(self, ec_id):
	#
	#	return RTC.RTC_OK
	
	#def onShutdown(self, ec_id):
	#
	#	return RTC.RTC_OK
	
	def onActivated(self, ec_id):
            self.Simulater.init_ODE()
	    return RTC.RTC_OK
	
	def onDeactivated(self, ec_id):
            self.Simulater.destroy()
	    return RTC.RTC_OK
	
	def onExecute(self, ec_id):
            
            dt = 0.01

            

            t = dt - (time.time() - self.Simulater.lasttime)
            if (t > 0):
                time.sleep(t)

            if self._TorqueIn.isNew():
                inport = self._TorqueIn.read()
                for i in range(0, len(inport.data)):
                    self.Simulater.joints[i].theta = inport.data[i]
            if self._DesFlagIn.isNew():
                self.Simulater.destroy()
                self.Simulater.init_ODE()
            
                
            self.Simulater.Pcontrol()
            self.Simulater.space.collide((self.Simulater.world,self.Simulater.contactgroup), self.Simulater.near_callback)
            self.Simulater.world.step(dt)
            self.Simulater.contactgroup.empty()

            print self.Simulater.bodies[0].getPosition()
            
            self.Simulater.lasttime = time.time()

            self._d_Pos.data = []
            self._d_Angle.data = []
            self._d_Pre.data = []
            self._d_Rot.data = []
            for i in range(0, self.Simulater.jointnum):
                self._d_Angle.data.append(self.Simulater.joints[i].getAngle)
            for i in range(0, self.Simulater.bodynum):
                for j in range(0, 9):
                    self._d_Rot.data.append(self.Simulater.bodies[i].getRotation()[j])
                for j in range(0, 3):
                    self._d_Pos.data.append(self.Simulater.bodies[0].getPosition()[j])
            for i in range(0, self.Simulater.datanum):
                self._d_Pre.data.append(self.Simulater.data[i])

            self._PosOut.write()
            self._RotOut.write()
            self._AngleOut.write()
            self._PreOut.write()

            
            
                    
        
	    return RTC.RTC_OK
	
	#def onAborting(self, ec_id):
	#
	#	return RTC.RTC_OK
	
	#def onError(self, ec_id):
	#
	#	return RTC.RTC_OK
	
	#def onReset(self, ec_id):
	#
	#	return RTC.RTC_OK
	
	#def onStateUpdate(self, ec_id):
	#
	#	return RTC.RTC_OK
	
	#def onRateChanged(self, ec_id):
	#
	#	return RTC.RTC_OK
	



class Simu:
    def __init__(self):
        self.bodynum = 0
        self.jointnum = 0
        self.bodies = []
        self.joints = []
        self.world = ode.World() #動力学の世界の創造
        self.space = ode.Space()
        self.lasttime = 0
        self.floor = ode.GeomPlane(self.space, (0,0,1), 0)
        self.contactgroup = ode.JointGroup()
        self.data =[]
        self.datanum = 0
        self.dnum = []
        self.world.setGravity((0,0,-0.2)) 
        self.world.setERP(1.0)
        self.world.setCFM(1E-5)
        
        
    

    def create_box(self, lx, ly, lz, px, py, pz, Mass):
        link = ode.Body(self.world)
        link.MASS = Mass
        link.lx = lx
        link.ly = ly
        link.lz = lz
        link.px = px
        link.py = py
        link.pz = pz
        link.shape = "box"

        self.bodies.append(link)
    

        mass = ode.Mass()
        mass.setBox(self.bodies[self.bodynum].MASS, self.bodies[self.bodynum].lx, self.bodies[self.bodynum].ly, self.bodies[self.bodynum].lz)
        link.setMass(mass)
        link.setPosition((self.bodies[self.bodynum].px, self.bodies[self.bodynum].py, self.bodies[self.bodynum].pz))
        link.boxsize = (self.bodies[self.bodynum].lx, self.bodies[self.bodynum].ly, self.bodies[self.bodynum].lz)
        geom = ode.GeomBox(self.space, self.bodies[self.bodynum].boxsize)
        geom.setBody(self.bodies[self.bodynum])
        self.bodynum += 1
        
    def create_sphere(self, r, px, py, pz, Mass):
        link = ode.Body(self.world)
        link.MASS = Mass
        link.r = r
        link.px = px
        link.py = py
        link.pz = pz
        link.shape = "sphere"

        self.bodies.append(link)
    

        mass = ode.Mass()
        mass.setSphere(self.bodies[self.bodynum].MASS, self.bodies[self.bodynum].r)
        link.setMass(mass)
        link.setPosition((self.bodies[self.bodynum].px, self.bodies[self.bodynum].py, self.bodies[self.bodynum].pz))
        geom = ode.GeomSphere(self.space, self.bodies[self.bodynum].r)
        geom.setBody(self.bodies[self.bodynum])
        self.bodynum += 1

    def create_cylinder(self, dn,  l, r ,px, py, pz, Mass):
        link = ode.Body(self.world)
        link.MASS = Mass
        link.l = l
        link.r = r
        link.px = px
        link.py = py
        link.pz = pz
        link.dn = dn
        link.shape = "cylinder"
        self.bodies.append(link)
        
        mass = ode.Mass()
        mass.setCylinder(self.bodies[self.bodynum].MASS, self.bodies[self.bodynum].dn, self.bodies[self.bodynum].r, self.bodies[self.bodynum].l)
        link.setMass(mass)
        link.setPosition((self.bodies[self.bodynum].px, self.bodies[self.bodynum].py, self.bodies[self.bodynum].pz))
        geom = ode.GeomCylinder(self.space, self.bodies[self.bodynum].r, self.bodies[self.bodynum].l)
        geom.setBody(self.bodies[self.bodynum])
        self.bodynum += 1
    def create_capsule(self, dn,  l, r, px, py, pz,Mass):
        link = ode.Body(self.world)
        link.MASS = Mass
        link.l = l
        link.r = r
        link.px = px
        link.py = py
        link.pz = pz
        link.dn = dn
        link.shape = "capsule"
        self.bodies.append(link)
    

        mass = ode.Mass()
        mass.setCappedCylinder(self.bodies[self.bodynum].MASS, self.bodies[self.bodynum].dn, self.bodies[self.bodynum].r, self.bodies[self.bodynum].l)
        link.setMass(mass)
        link.setPosition((self.bodies[self.bodynum].px, self.bodies[self.bodynum].py, self.bodies[self.bodynum].pz))
        geom = ode.GeomCapsule(self.space, self.bodies[self.bodynum].r, self.bodies[self.bodynum].l)
        geom.setBody(self.bodies[self.bodynum])
        self.bodynum += 1
        
    def create_Fix_joint(self, link1, link2):

        joint = ode.FixedJoint(self.world)
        joint.name = "FIXED"
        self.joints.append(joint)
        self.joints[self.jointnum].attach(link1, link2)
        self.joints[self.jointnum].setFixed()
        self.jointnum += 1
    
    def create_Hinge_joint(self, link1, link2, c_x, c_y, c_z, axis_x, axis_y, axis_z, control):

        joint = ode.HingeJoint(self.world)
        joint.name = "HINGE"
        joint.control = control
        joint.theta = 0
        self.joints.append(joint)
        self.joints[self.jointnum].attach(link1, link2)
        self.joints[self.jointnum].setAnchor( (c_x,c_y,c_z) )
        self.joints[self.jointnum].setAxis((axis_x,axis_y,axis_z))
        self.joints[self.jointnum].setParam(ode.ParamFudgeFactor, 0.1)
        self.joints[self.jointnum].setParam(ode.ParamSuspensionCFM,0.015)
        self.joints[self.jointnum].setParam(ode.ParamSuspensionERP,0.8)
        self.jointnum += 1

    def create_Slider_joint(self, link1, link2, axis_x, axis_y, axis_z):

        joint = ode.SliderJoint(self.world)
        joint.name = "SLIDER"
        joint.theta = 0
        self.joints.append(joint)
        self.joints[self.jointnum].attach(link1, link2)
        self.joints[self.jointnum].setAxis((axis_x,axis_y,axis_z))
        self.joints[self.jointnum].setParam(ode.ParamHiStop, 0.1)
        self.joints[self.jointnum].setParam(ode.ParamLoStop, -0.1)
        self.joints[self.jointnum].setParam(ode.ParamFudgeFactor, 0.1)
        self.jointnum += 1

    def create_Ball_joint(self, link1, link2, c_x, c_y, c_z):

        joint = ode.BallJoint(self.world)
        joint.name = "BALL"
        joint.theta = 0
        self.joints.append(joint)
        self.joints[self.jointnum].attach(link1, link2)
        self.joints[self.jointnum].setAnchor( (c_x,c_y,c_z) )
        self.joints[self.jointnum].setParam(ode.ParamFudgeFactor, 0.1)
        self.jointnum += 1
        
    def create_Hinge2_joint(self, link1, link2, c_x, c_y, c_z, axis_x1, axis_y1, axis_z1, axis_x2, axis_y2, axis_z2):

        joint = ode.Hinge2Joint(self.world)
        joint.name = "HINGE2"
        joint.theta = 0
        joint.theta2 = 0
        self.joints.append(joint)
        self.joints[self.jointnum].attach(link1, link2)
        self.joints[self.jointnum].setAnchor( (c_x,c_y,c_z) )
        self.joints[self.jointnum].setAxis1((axis_x1,axis_y1,axis_z1))
        self.joints[self.jointnum].setAxis2((axis_x2,axis_y2,axis_z2))
        self.joints[self.jointnum].setParam(ode.ParamFudgeFactor, 0.1)
        self.jointnum += 1
        
    def create_Universal_joint(self, link1, link2, c_x, c_y, c_z, axis_x1, axis_y1, axis_z1, axis_x2, axis_y2, axis_z2):

        joint = ode.UniversalJoint(self.world)
        joint.name = "UNIVERSAL"
        joint.theta = 0
        joint.theta2 = 0
        self.joints.append(joint)
        self.joints[self.jointnum].attach(link1, link2)
        self.joints[self.jointnum].setAnchor( (c_x,c_y,c_z) )
        self.joints[self.jointnum].setAxis1((axis_x1,axis_y1,axis_z1))
        self.joints[self.jointnum].setAxis2((axis_x2,axis_y2,axis_z2))
        self.joints[self.jointnum].setParam(ode.ParamFudgeFactor, 0.1)
        self.jointnum += 1

    def create_Amotor(self, link1, link2, axis_x, axis_y, axis_z, anum, rel, num_axis):

        joint = ode.AMotor(self.world)
        joint.name = "AMOTOR"
        joint.theta = 0
        joint.anum = anum
        self.joints.append(joint)
        self.joints[self.jointnum].attach(link1, link2)
        #self.joints[self.jointnum].setMode(AMotorUser)
        self.joints[self.jointnum].setNumAxes(num_axis)
        self.joints[self.jointnum].setAxis(anum, rel, (axis_x,axis_y,axis_z))
        self.joints[self.jointnum].setParam(ode.ParamFudgeFactor, 0.1)
        self.jointnum += 1
        
    def Pcontrol(self):
        kp = 5.0
        kv = 3.0
        k = 10
        fMax = 200.0
        for i in range(0, self.jointnum):
            if self.joints[i].name == "HINGE":
                if self.joints[i].control == "T":
                    tmp = self.joints[i].getAngle()
                    z = self.joints[i].theta - tmp
                    d = self.joints[i].getAngleRate()
                    self.joints[i].setParam(ode.ParamVel, k*z)
                    self.joints[i].setParam(ode.ParamFMax, fMax)
                    #self.joints[i].addTorque(kp*z-kv*d)
            if self.joints[i].name == "SLIDER":
                tmp = self.joints[i].getPosition()
                z = self.joints[i].theta - tmp
                d = self.joints[i].getPositionRate()
                self.joints[i].setParam(ode.ParamVel, k*z)
                self.joints[i].setParam(ode.ParamFMax, fMax)
                #self.joints[i].addForce(kp*z-kv*d)
            if self.joints[i].name == "BALL":
                pass
            if self.joints[i].name == "HINGE2":
                tmp = self.joints[i].getAngle1()
                z = self.joints[i].theta - tmp
                self.joints[i].setParam(ode.ParamVel, 0)
                self.joints[i].setParam(ode.ParamFMax, fMax)
                self.joints[i].setParam(ode.ParamSuspensionERP, 0.4)
                self.joints[i].setParam(ode.ParamSuspensionCFM, 0.8)
            if self.joints[i].name == "UNIVERSAL":
                self.joints[i].setParam(ode.ParamVel, 0)
                self.joints[i].setParam(ode.ParamFMax, fMax)
                self.joints[i].setParam(ode.ParamVel2, 0)
                self.joints[i].setParam(ode.ParamFMax2, fMax)
            if self.joints[i].name == "AMOTOR":
                tmp = self.joints[i].getAngle(self.joints[i].anum)
                z = self.joints[i].theta - tmp
                #d = self.joints[i].getAngleRate(self.joints[i].anum)
                #self.joints[i].setParam(ode.ParamVel, k*z)
                #self.joints[i].setParam(ode.ParamFMax, fMax)
                #self.joints[i].addTorque(kp*z-kv*d)
    
                
            
    def near_callback(self, args, geom1, geom2):

        contacts = ode.collide(geom1, geom2)

        self.world,self.contactgroup = args
        if geom1 == self.floor or geom2 == self.floor:
            for c in contacts:
                c.setBounce(0.2)
                c.setMu(5000)
                j = ode.ContactJoint(self.world, self.contactgroup, c)
                j.attach(geom1.getBody(), geom2.getBody())
                
    def init_force_data(self, num):
        self.joints[num].setFeedback(True)
        d = []
        self.data.append(d)
        self.data[self.datanum] = self.joints[num].getFeedback()
        self.dnum.append(num)
        self.datanum += 1
    def get_force_data(self):
        for i in range(0, self.datanum):
            d = self.joints[self.dnum[i]].getFeedback()
            self.data[i] = d
        
    def init_ODE(self):
        
        self.world.setGravity((0,0,-9.8))
        self.floor = ode.GeomPlane(self.space, (0,0,1), 0)
        
        self.create_box(0.2, 0.7, 0.2, 0.0, 0.0, 0.2, 10.0)
        """self.create_box(0.2, 0.7, 0.2, 0.0, 0.7, 0.2, 10.0)
        self.create_box(0.2, 0.7, 0.2, 0.0, 1.4, 0.2, 10.0)
        self.create_box(0.2, 0.7, 0.2, 0.0, 2.1, 0.2, 10.0)
        self.create_box(0.2, 0.7, 0.2, 0.0, 2.8, 0.2, 10.0)
        self.create_box(0.2, 0.7, 0.2, 0.0, 3.5, 0.2, 10.0)
        self.create_sphere(0.1, -0.2, 0.0, 0.19, 3.0)
        self.create_sphere(0.1, 0.2, 0.0, 0.19, 3.0)
        self.create_sphere(0.1, -0.2, 0.7, 0.19, 3.0)
        self.create_sphere(0.1, 0.2, 0.7, 0.19, 3.0)
        self.create_sphere(0.1, -0.2, 1.4, 0.19, 3.0)
        self.create_sphere(0.1, 0.2, 1.4, 0.19, 3.0)
        self.create_sphere(0.1, -0.2, 2.1, 0.19, 3.0)
        self.create_sphere(0.1, 0.2, 2.1, 0.19, 3.0)
        self.create_sphere(0.1, -0.2, 2.8, 0.19, 3.0)
        self.create_sphere(0.1, 0.2, 2.8, 0.19, 3.0)
        self.create_sphere(0.1, -0.2, 3.5, 0.19, 3.0)
        self.create_sphere(0.1, 0.2, 3.5, 0.19, 3.0)
        self.create_Hinge_joint(self.bodies[0], self.bodies[1], 0, 0.35, 0.3, 0, 0, 1, "T")
        self.create_Hinge_joint(self.bodies[1], self.bodies[2], 0, 1.05, 0.3, 0, 0, 1, "T")
        self.create_Hinge_joint(self.bodies[2], self.bodies[3], 0, 1.75, 0.3, 0, 0, 1, "T")
        self.create_Hinge_joint(self.bodies[3], self.bodies[4], 0, 2.45, 0.3, 0, 0, 1, "T")
        self.create_Hinge_joint(self.bodies[4], self.bodies[5], 0, 3.15, 0.3, 0, 0, 1, "T")
        self.create_Hinge_joint(self.bodies[0], self.bodies[6], 0.1, 0.00, 0.19, 1, 0, 0, "F")
        self.create_Hinge_joint(self.bodies[0], self.bodies[7], -0.1, 0.00, 0.19, 1, 0, 0, "F")
        self.create_Hinge_joint(self.bodies[1], self.bodies[8], 0.1, 0.7, 0.19, 1, 0, 0, "F")
        self.create_Hinge_joint(self.bodies[1], self.bodies[9], -0.1, 0.7, 0.19, 1, 0, 0, "F")
        self.create_Hinge_joint(self.bodies[2], self.bodies[10], 0.1, 1.4, 0.19, 1, 0, 0, "F")
        self.create_Hinge_joint(self.bodies[2], self.bodies[11], -0.1, 1.4, 0.19, 1, 0, 0, "F")
        self.create_Hinge_joint(self.bodies[3], self.bodies[12], 0.1, 2.1, 0.19, 1, 0, 0, "F")
        self.create_Hinge_joint(self.bodies[3], self.bodies[13], -0.1, 2.1, 0.19, 1, 0, 0, "F")
        self.create_Hinge_joint(self.bodies[4], self.bodies[14], 0.1, 2.8, 0.19, 1, 0, 0, "F")
        self.create_Hinge_joint(self.bodies[4], self.bodies[15], -0.1, 2.8, 0.19, 1, 0, 0, "F")
        self.create_Hinge_joint(self.bodies[5], self.bodies[16], 0.1, 3.5, 0.19, 1, 0, 0, "F")
        self.create_Hinge_joint(self.bodies[5], self.bodies[17], -0.1, 3.5, 0.19, 1, 0, 0, "F")"""
    def destroy(self):
        self.bodies = []
        self.joints = []
        self.jointnum = 0
        self.bodynum = 0
        self.world = ode.World()
        self.space = ode.Space()
        self.step = 0
        self.contactgroup = ode.JointGroup()
        self.data =[]
        self.datanum = 0
        self.dnum = []
        self.posdata = []
        self.posdatanum = 0

def MyModuleInit(manager):
    profile = OpenRTM.Properties(defaults_str=pySimu_spec)
    manager.registerFactory(profile,
                            pySimu,
                            OpenRTM.Delete)

    # Create a component
    comp = manager.createComponent("pySimu")



def main():
	mgr = OpenRTM.Manager.init(len(sys.argv), sys.argv)
	mgr.setModuleInitProc(MyModuleInit)
	mgr.activateManager()
	mgr.runManager()

if __name__ == "__main__":
	main()

